@@ -21,23 +21,73 @@
 # define __CDIO_CONFIG_H__ 1
 #endif
 
+#ifdef HAVE_STDIO_H
 #include <stdio.h>
+#endif
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
-
+#endif
 #ifdef HAVE_UNISTD_H
-#include <unistd.h> 
-#endif /*HAVE_UNISTD_H*/
-
+#include <unistd.h>
+#endif
+#ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
+#endif
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
+#include <ctype.h>
 
 #include <cdio/logging.h>
 #include <cdio/util.h>
 #include "_cdio_stream.h"
 #include "_cdio_stdio.h"
+#include "cdio_assert.h"
 
-static const char _rcsid[] = "$Id: _cdio_stdio.c,v 1.6 2008/04/22 15:29:11 karl Exp $";
+/* On 32 bit platforms, fseek can only access streams of 2 GB or less.
+   Prefer fseeko/fseeko64, that take a 64 bit offset when LFS is enabled */
+#if defined(HAVE_FSEEKO64) && defined(_FILE_OFFSET_BITS) && (_FILE_OFFSET_BITS == 64)
+#define CDIO_FSEEK fseeko64
+#elif defined(HAVE_FSEEKO)
+#define CDIO_FSEEK fseeko
+#else
+#define CDIO_FSEEK fseek
+#endif
+
+/* Windows' fopen is not UTF-8 compliant, so we use our own */
+#if defined(_WIN32)
+#include <cdio/utf8.h>
+#define CDIO_FOPEN fopen_utf8
+#else
+#define CDIO_FOPEN fopen
+#endif
+
+/* Use _stati64 if needed, on platforms that don't have transparent LFS support */
+#if defined(HAVE__STATI64) && defined(_FILE_OFFSET_BITS) && (_FILE_OFFSET_BITS == 64)
+#define CDIO_STAT_STRUCT _stati64
+#if defined(_WIN32)
+/* Once again, use our own UTF-8 compliant version */
+static inline int _stati64_utf8(const char *path, struct _stati64 *buffer) {
+  int ret;
+  wchar_t* wpath = cdio_utf8_to_wchar(path);
+  ret = _wstati64(wpath, buffer);
+  cdio_free(wpath);
+  return ret;
+}
+#define CDIO_STAT_CALL _stati64_utf8
+#else
+#define CDIO_STAT_CALL _stati64
+#endif
+#else
+#define CDIO_STAT_STRUCT stat
+#define CDIO_STAT_CALL stat
+#endif
+
+#define _STRINGIFY(a) #a
+#define STRINGIFY(a) _STRINGIFY(a)
 
 #define CDIO_STDIO_BUFSIZE (128*1024)
 
@@ -49,11 +99,11 @@ typedef struct {
 } _UserData;
 
 static int
-_stdio_open (void *user_data) 
+_stdio_open (void *user_data)
 {
   _UserData *const ud = user_data;
-  
-  if ((ud->fd = fopen (ud->pathname, "rb")))
+
+  if ((ud->fd = CDIO_FOPEN (ud->pathname, "rb")))
     {
       ud->fd_buf = calloc (1, CDIO_STDIO_BUFSIZE);
       setvbuf (ud->fd, ud->fd_buf, _IOFBF, CDIO_STDIO_BUFSIZE);
@@ -69,7 +119,7 @@ _stdio_close(void *user_data)
 
   if (fclose (ud->fd))
     cdio_error ("fclose (): %s", strerror (errno));
- 
+
   ud->fd = NULL;
 
   free (ud->fd_buf);
@@ -87,14 +137,14 @@ _stdio_free(void *user_data)
     free(ud->pathname);
 
   if (ud->fd) /* should be NULL anyway... */
-    _stdio_close(user_data); 
+    _stdio_close(user_data);
 
   free(ud);
 }
 
-/*! 
-  Like fseek(3) and in fact may be the same.
-  
+/*!
+  Like fseek/fseeko(3) and in fact may be the same.
+
   This  function sets the file position indicator for the stream
   pointed to by stream.  The new position, measured in bytes, is obtained
   by  adding offset bytes to the position specified by whence.  If whence
@@ -103,24 +153,33 @@ _stdio_free(void *user_data)
   respectively.  A successful call to the fseek function clears the end-
   of-file indicator for the stream and undoes any effects of the
   ungetc(3) function on the same stream.
-  
+
   @return upon successful completion, DRIVER_OP_SUCCESS, else,
   DRIVER_OP_ERROR is returned and the global variable errno is set to
   indicate the error.
 */
-static driver_return_code_t 
-_stdio_seek(void *p_user_data, long i_offset, int whence)
+static int
+_stdio_seek(void *p_user_data, off_t i_offset, int whence)
 {
   _UserData *const ud = p_user_data;
+  int ret;
+#if !defined(HAVE_FSEEKO) && !defined(HAVE_FSEEKO64)
+  /* Detect if off_t is lossy-truncated to long to avoid data corruption */
+  if ( (sizeof(off_t) > sizeof(long)) && (i_offset != (off_t)((long)i_offset)) ) {
+    cdio_error ( STRINGIFY(CDIO_FSEEK) " (): lossy truncation detected!");
+    errno = EFBIG;
+    return DRIVER_OP_ERROR;
+  }
+#endif
 
-  if ( (i_offset=fseek (ud->fd, i_offset, whence)) ) {
-    cdio_error ("fseek (): %s", strerror (errno));
+  if ( (ret=CDIO_FSEEK (ud->fd, i_offset, whence)) ) {
+    cdio_error ( STRINGIFY(CDIO_FSEEK) " (): %s", strerror (errno));
   }
 
-  return i_offset;
+  return ret;
 }
 
-static long int
+static off_t
 _stdio_stat(void *p_user_data)
 {
   const _UserData *const ud = p_user_data;
@@ -130,30 +189,30 @@ _stdio_stat(void *p_user_data)
 
 /*!
   Like fread(3) and in fact is about the same.
-  
+
   DESCRIPTION:
   The function fread reads nmemb elements of data, each size bytes long,
   from the stream pointed to by stream, storing them at the location
   given by ptr.
-  
+
   RETURN VALUE:
   return the number of items successfully read or written (i.e.,
   not the number of characters).  If an error occurs, or the
   end-of-file is reached, the return value is a short item count
   (or zero).
-  
+
   We do not distinguish between end-of-file and error, and callers
   must use feof(3) and ferror(3) to determine which occurred.
   */
-static long
-_stdio_read(void *user_data, void *buf, long int count)
+static ssize_t
+_stdio_read(void *user_data, void *buf, size_t count)
 {
   _UserData *const ud = user_data;
-  long read;
+  long read_count;
 
-  read = fread(buf, 1, count, ud->fd);
+  read_count = fread(buf, 1, count, ud->fd);
 
-  if (read != count)
+  if (read_count != count)
     { /* fixme -- ferror/feof */
       if (feof (ud->fd))
         {
@@ -169,7 +228,7 @@ _stdio_read(void *user_data, void *buf, long int count)
         cdio_debug ("fread (): short read and no EOF?!?");
     }
 
-  return read;
+  return read_count;
 }
 
 /*!
@@ -187,18 +246,29 @@ cdio_stdio_new(const char pathname[])
   CdioDataSource_t *new_obj = NULL;
   cdio_stream_io_functions funcs = { NULL, NULL, NULL, NULL, NULL, NULL };
   _UserData *ud = NULL;
-  struct stat statbuf;
-  
-  if (stat (pathname, &statbuf) == -1) 
+  struct CDIO_STAT_STRUCT statbuf;
+  char* pathdup;
+
+  if (pathname == NULL)
+    return NULL;
+
+  /* MinGW may require a translated path */
+  pathdup = _cdio_strdup_fixpath(pathname);
+  if (pathdup == NULL)
+    return NULL;
+
+  if (CDIO_STAT_CALL (pathdup, &statbuf) == -1)
     {
-      cdio_warn ("could not retrieve file info for `%s': %s", 
-                 pathname, strerror (errno));
+      cdio_warn ("could not retrieve file info for `%s': %s",
+                 pathdup, strerror (errno));
+      cdio_free(pathdup);
       return NULL;
     }
 
   ud = calloc (1, sizeof (_UserData));
+  cdio_assert (ud != NULL);
 
-  ud->pathname = strdup(pathname);
+  ud->pathname = pathdup;
   ud->st_size  = statbuf.st_size; /* let's hope it doesn't change... */
 
   funcs.open   = _stdio_open;
@@ -214,7 +284,7 @@ cdio_stdio_new(const char pathname[])
 }
 
 
-/* 
+/*
  * Local variables:
  *  c-file-style: "gnu"
  *  tab-width: 8
